import os

from thundra import constants


def get_thundra_apikey():
    return os.environ['thundra_apiKey'] if 'thundra_apiKey' in os.environ else None


def is_thundra_disabled():
    should_thundra_disabled = os.environ['thundra_disable'] if 'thundra_disable' in os.environ else 'false'
    return should_disable(should_thundra_disabled)


def should_disable(disable_by_env, disable_by_param=False):
    if disable_by_env == 'true':
        return True
    elif disable_by_env == 'false':
        return False
    else:
        return disable_by_param

def is_thundra_trace_disabled():
    return os.environ['thundra_trace_disable'] if 'thundra_trace_disable' in os.environ else ''


def is_thundra_lambda_publish_cloudwatch_enabled():
    return os.environ['thundra_lambda_publish_cloudwatch_enable'] if 'thundra_lambda_publish_cloudwatch_enable' in os.environ else 'false'


def is_thundra_lambda_audit_request_skipped():
    return os.environ['thundra_lambda_audit_request_skip'] if 'thundra_lambda_audit_request_skip' in os.environ else ''


def is_thundra_lambda_audit_response_skipped():
    return os.environ['thundra_lambda_audit_response_skip'] if 'thundra_lambda_audit_response_skip' in os.environ else ''


def get_thundra_application_profile():
    return os.environ['thundra_applicationProfile'] if 'thundra_applicationProfile' in os.environ else ''


def get_aws_lambda_log_stream():
    return os.environ['AWS_LAMBDA_LOG_STREAM_NAME'] if 'AWS_LAMBDA_LOG_STREAM_NAME' in os.environ else None


def get_aws_lambda_function_version():
    return os.environ['AWS_LAMBDA_FUNCTION_VERSION'] if 'AWS_LAMBDA_FUNCTION_VERSION' in os.environ else ""


def get_aws_region():
    return os.environ['AWS_REGION'] if 'AWS_REGION' in os.environ else ""


def format_date(date_to_be_formatted):
    date = date_to_be_formatted.strftime(constants.DATE_FORMAT)
    # make ms 3 digits
    ms = date[20:-6]
    formatted_date = date.replace(ms, ms[:-3])
    return formatted_date


from functools import wraps
import time

from thundra import constants
from thundra.plugins.trace.trace_plugin import TracePlugin
from thundra.reporter import Reporter

import thundra.utils as utils


class Thundra:
    def __init__(self, api_key=None, disable_trace=False, request_skip=False, response_skip=False):

        constants.REQUEST_COUNT = 0

        self.plugins = []
        api_key_from_environment_variable = utils.get_thundra_apikey()
        self.api_key = api_key_from_environment_variable if api_key_from_environment_variable is not None else api_key
        if self.api_key is None:
            raise Exception('Please set thundra_apiKey from environment variables in order to use Thundra')
        self.data = {}

        disable_trace_by_env = utils.is_thundra_trace_disabled()
        if not utils.should_disable(disable_trace_by_env, disable_trace):
            self.plugins.append(TracePlugin())

        audit_request_skip_by_env = utils.is_thundra_lambda_audit_request_skipped()
        self.data['request_skipped'] = utils.should_disable(audit_request_skip_by_env, request_skip)

        audit_response_skip_by_env = utils.is_thundra_lambda_audit_response_skipped()
        self.response_skipped = utils.should_disable(audit_response_skip_by_env, response_skip)

        self.reporter = Reporter(self.api_key)

    def __call__(self, original_func):

        should_disable_thundra = utils.is_thundra_disabled()
        if should_disable_thundra:
            return original_func

        @wraps(original_func)
        def wrapper(event, context):
            if self.checkAndHandleWarmupRequest(event):
                constants.REQUEST_COUNT += 1
                return None
            self.data['reporter'] = self.reporter
            self.data['event'] = event
            self.data['context'] = context
            self.execute_hook('before:invocation', self.data)
            try:
                response = original_func(event, context)
                if self.response_skipped is not True:
                    self.data['response'] = response
            except Exception as e:
                self.data['error'] = e
                self.execute_hook('after:invocation', self.data)
                self.reporter.send_report()
                raise e
            self.execute_hook('after:invocation', self.data)
            self.reporter.send_report()
            return response

        return wrapper

    call = __call__

    def execute_hook(self, name, data):
        [plugin.hooks[name](data) for plugin in self.plugins if hasattr(plugin, 'hooks') and name in plugin.hooks]

    def checkAndHandleWarmupRequest(self, event):

        # Check whether it is empty request which is used as default warmup request
        if (not event):
            print("Received warmup request as empty message. " +
                  "Handling with 100 milliseconds delay ...")
            time.sleep(0.1)
            return True
        else:
            if (isinstance(event, str)):
                # Check whether it is warmup request
                if (event.startswith('#warmup')):
                    delayTime = 100
                    args = event[len('#warmup'):].strip().split()
                    # Warmup messages are in '#warmup wait=<waitTime>' format
                    # Iterate over all warmup arguments
                    for arg in args:
                        argParts = arg.split('=')
                        # Check whether argument is in key=value format
                        if (len(argParts) == 2):
                            argName = argParts[0]
                            argValue = argParts[1]
                            # Check whether argument is "wait" argument
                            # which specifies extra wait time before returning from request
                            if (argName == 'wait'):
                                waitTime = int(argValue)
                                delayTime += waitTime
                    print("Received warmup request as warmup message. " +
                          "Handling with " + str(delayTime) + " milliseconds delay ...")
                    time.sleep(delayTime / 1000)
                    return True
            return False





#from datetime import datetime, timezone
import time
import uuid

import thundra.utils as utils
from thundra import constants


class TracePlugin:

    data_format_version = '1.1'
    IS_COLD_START = True

    def __init__(self):
        self.hooks = {
            'before:invocation': self.before_invocation,
            'after:invocation': self.after_invocation
        }
        self.start_time = 0
        self.end_time = 0
        aws_lambda_log_stream = utils.get_aws_lambda_log_stream()
        self.log_stream = aws_lambda_log_stream.split("]")[1] if aws_lambda_log_stream is not None else ''
        self.application_version = utils.get_aws_lambda_function_version()
        self.application_profile = utils.get_thundra_application_profile()
        self.aws_region = utils.get_aws_region()
        self.trace_data = {}

    def before_invocation(self, data):

        if constants.REQUEST_COUNT > 0:
            TracePlugin.IS_COLD_START = False
        context = data['context']
        event = data['event']

        context_id = str(uuid.uuid4())
        #self.start_time = datetime.now(timezone.utc)
        #formatted_date = utils.format_date(self.start_time)
        self.start_time = time.time() * 1000
        self.trace_data = {
            'id': str(uuid.uuid4()),
            'applicationName': context.function_name,
            'applicationId': self.log_stream,
            'applicationVersion': self.application_version,
            'applicationProfile': self.application_profile,
            'applicationType': 'python',
            'duration': None,
            'startTimestamp': int(self.start_time),
            'endTime': None,
            'errors': [],
            'thrownError': None,
            'contextType': 'ExecutionContext',
            'contextName': context.function_name,
            'contextId': context_id,
            'auditInfo': {
                'contextName': context.function_name,
                'id': context_id,
                'openTimestamp': int(self.start_time),
                'closeTime': None,
                'errors': [],
                'thrownError': None
            },
            'properties': {
                'request': event if data['request_skipped'] is False else None,
                'response': {},
                'coldStart': 'true' if TracePlugin.IS_COLD_START else 'false',
                'functionRegion': self.aws_region,
                'functionMemoryLimitInMB': context.memory_limit_in_mb
            }

        }
        TracePlugin.IS_COLD_START = False

    def after_invocation(self, data):
        if 'error' in data:
            error = data['error']
            error_type = type(error)
            exception = {
                'errorType': error_type.__name__,
                'errorMessage': str(error),
                'args': error.args,
                'cause': error.__cause__
            }
            self.trace_data['errors'].append(error_type.__name__)
            self.trace_data['thrownError'] = error_type.__name__
            self.trace_data['auditInfo']['errors'].append(exception)
            self.trace_data['auditInfo']['thrownError'] = exception

        if 'response' in data:
            self.trace_data['properties']['response'] = data['response']
        #self.end_time = datetime.now(timezone.utc)
        self.end_time = time.time() * 1000
        duration = self.end_time - self.start_time
        #duration_in_ms = duration * 1000
        #duration_in_ms = duration.total_seconds() * 1000
        self.trace_data['duration'] = int(duration)
        #formatted_date = utils.format_date(self.end_time)
        self.trace_data['endTimestamp'] = int(self.end_time)
        self.trace_data['auditInfo']['closeTimestamp'] = int(self.end_time)

        reporter = data['reporter']
        report_data = {
            'apiKey': reporter.api_key,
            'type': 'AuditData',
            'dataFormatVersion': TracePlugin.data_format_version,
            'data': self.trace_data
        }
        reporter.add_report(report_data)






import json
import os

try:
    import requests
except ImportError:
    from botocore.vendored import requests

from thundra import constants
from thundra.utils import is_thundra_lambda_publish_cloudwatch_enabled


class Reporter:
    def __init__(self, api_key):
        if api_key is not None:
            self.api_key = api_key
        else:
            raise Exception('Please set an api key')
        self.reports = []

    def add_report(self, report):
        if is_thundra_lambda_publish_cloudwatch_enabled() == 'true':
            print(json.dumps(report))
        else:
            self.reports.append(report)

    def send_report(self):
        request_url = constants.HOST + constants.PATH
        headers = {
            'Content-Type': 'application/json',
            'Authorization': 'ApiKey ' + self.api_key
        }
        response = requests.post(request_url, headers=headers, data=json.dumps(self.reports))
        return response



from setuptools import setup, find_packages

setup(name='thundra',
      version='0.0.12',
      description='Thundra Python agent',
      url='https://github.com/thundra-io/thundra-lambda-agent-python',
      author='Thundra',
      author_email='support@thundra.io',
      packages=find_packages(exclude=('tests', 'tests.*',)),
      python_requires='>=3',
      zip_safe=True,
      )